#!/usr/bin/perl -w
use strict;
use warnings;

use Image::Magick;
use DBI;
use IO::Socket::IP;
use POSIX;
use Proc::Daemon;
#use Data::Dumper;

Proc::Daemon::Init;
my $continue = 1;
$SIG{TERM} = sub{ $continue = 0 };

use constant FONT => "/usr/share/fonts/truetype/dejavu/DejaVuSans.ttf";

# auto-flush on socket
$| = 1;

my $socket = IO::Socket::IP->new(
    LocalHost => '0.0.0.0',
    LocalPort => '1056',
    Proto => 'tcp',
    Listen => 5,
    Reuse => 1
);
die "Cannot create socket $!\n" unless $socket;

print "\nServer started, waiting for connection on port 1056.\n\n";

my $client_cache = {};


while( $continue ) {
    my $client_socket = $socket->accept;

# der laesst sockets offen, is mist!
#    my $data = $client_socket->getline();
#    $data =~ s/[^a-z]//g;

    my $data = "";
    $client_socket->recv( $data, 4 );
    $data =~ s/[^a-z]//g;

    # get information about a newly connected client
    my $client_address = $client_socket->peerhost();
    my $client_port = $client_socket->peerport();
    print "connection from $client_address:$client_port\n";

    # Update-Command will look into cache and send no data if
    # picture didn't change since last update
    if( $data eq "updt" ) {
        my $send_img = 0;
        if( defined( $client_cache->{$client_address} ) ) {
            if( time - $client_cache->{$client_address}->{"timestamp"} gt 60 ) {
                # Cache is outdated, create new one...
                $client_cache->{$client_address} = clientCache();
                $send_img = 1;
            } else {
                # Check if cache is still up to date...
                if(
                    $client_cache->{$client_address}->{"temperatur"} ne getTemp() ||
                    $client_cache->{$client_address}->{"rh"} ne getRh() ||
                    $client_cache->{$client_address}->{"zeit"} ne getZeit()
                ) {
                    # Cache and image has to be renewed...
                    $client_cache->{$client_address} = clientCache();
                    $send_img = 1;
                } else {
                    # Data did not change...
                    $client_socket->send( "" );
                }
            }
        } else {
            # Client never seen before, create cache and image
            $client_cache->{$client_address} = clientCache();
            $send_img = 1;
        }

        if( $send_img == 1 ) {
            print "Send image.\n";
            my $image = clocktemp(
                $client_cache->{$client_address}->{"zeit"},
                $client_cache->{$client_address}->{"temperatur"},
                $client_cache->{$client_address}->{"rh"}
            );
            $client_socket->send( $image );
        } else {
            print "Image-data did not change for this client.\n";
        }

    # Pull-Command will send a picture right away
    } elsif( $data eq "pull" ) {
        print "Send image (pull).\n";
        my $image = clocktemp( getZeit(), getTemp(), getRh() );
        $client_socket->send( $image );

    # Graph-Command get's temp-chart of last 128 minutes
    } elsif( $data eq "grph" ) {
        print "Send TempGraph.\n";
        my $image = createPlot();
        $client_socket->send( $image );
    }

    shutdown( $client_socket, 1 );
    print "\n";

    # Clean up client_cache (delete old data)...
    foreach my $key( keys %{$client_cache} ) {
        if( time - $client_cache->{$key}->{"timestamp"} gt 60 ) {
            print "Removing $key from client_cache - outdated.\n\n";
            delete $client_cache->{$key};
        }
    }
}
$socket->close();

sub connectTemp {
    my $dbh = DBI->connect('dbi:mysql:temperatur', 'USER', 'PASS' )
        or die( "Cannot connect to Temp-DB!\n");
    return( $dbh );
}

sub connectRh {
    my $dbh = DBI->connect('dbi:mysql:luftfeuchtigkeit', 'USER', 'PASS' )
        or die( "Cannot connect to Rh-DB!\n");
    return( $dbh );
}

sub getTemp {
    my $dbh = connectTemp();
    my $sth = $dbh->prepare( "SELECT `T3` FROM `t_sensor` ORDER BY `timestmp` DESC LIMIT 1" );
    $sth->execute
        or die "SQL Error: $DBI::errstr\n";

    my @row = $sth->fetchrow_array;
    $dbh->disconnect;

    return( $row[0] / 10 );
}

sub getRh {
    my $dbh = connectRh();
    my $sth = $dbh->prepare( "SELECT `RH` FROM `sensor_data` ORDER BY `timestmp` DESC LIMIT 1 ");
    $sth->execute
        or die "SQL Error: $DBI::errstr\n";

    my @row = $sth->fetchrow_array;
    $dbh->disconnect;

    return( $row[0] );
}

sub getTempGraph {
    my $dbh = connectTemp();
    my $sth = $dbh->prepare( "SELECT `T3` FROM `t_sensor` ORDER BY `timestmp` DESC LIMIT 128" );
    $sth->execute
        or die "SQL Error: $DBI::errstr\n";

    my %res;

    $res{'low'} = 0;	# Make sure lowest and highest Temp is at least 0°C
    $res{'high'} = 0;	# we draw a 0-line later...
    $res{'temp'} = ();

    my @row;
    while( @row = $sth->fetchrow_array ) {
        push @{$res{'temp'}}, $row[0] / 10;
        if( $row[0] / 10 > $res{'high'} ) {
            $res{'high'} = $row[0] / 10;
        }
        if( $row[0] / 10 < $res{'low'} ) {
            $res{'low'} = $row[0] / 10;
        }
    }
    $dbh->disconnect;

    return( \%res );
}

sub getZeit {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();
    my $zeit;
    if( length( $hour ) == 1 ) {
        $zeit = "0";
    }
    $zeit .= $hour.":";

    if( length( $min ) == 1 ) {
        $zeit .= "0";
    }
    $zeit .= $min." Uhr";

    return( $zeit );
}

sub yscale {
    my( $y, $low, $high ) = @_;
    my $abs = abs( $high - $low );
    my $scal = 42 / $abs; # Our graph's gonna be 42 pixels high...

    # Lowest temperature will be at 50 pixels
    return( floor( 50 - ( $y + abs( $low ) ) * $scal ) );
}

sub clientCache {
    my %cache;
    $cache{"timestamp"} = time;
    $cache{"temperatur"} = getTemp();
    $cache{"rh"} = getRh();
    $cache{"zeit"} = getZeit();

    return( \%cache );
}

sub createImg {
    my $img = Image::Magick->new();
    $img->Set( size => "128x64" );
    #$img->Quantize(colors => 2);
    $img->ReadImage( "xc:white" );
    $img->Set( monochrome => 'True', type => "bilevel" );
    $img->Set( antialias => 'False' );

    return( $img )
}

sub clocktemp {
    my( $zeit, $temperatur, $rh ) = @_;

    my $img = createImg();

    $img->Annotate(
        font => FONT,
        geometry => "+0+16",
        fill => 'black',
        text => $zeit,
        pointsize => 16
    );

#$temperatur = "32.4";

    $img->Annotate(
        font => FONT,
        gravity => "north",
        geometry => "+0+18",
        size => "128x",
        fill => 'black',
        text => $temperatur." °C",
        pointsize => 28
    );

    $img->Annotate(
        font => FONT,
        gravity => "north",
        geometry => "+0+48",
        size => "128x",
        fill => 'black',
        text => "φ ".$rh." %",
        pointsize => 16
    );

#    $img->Write( filename => "img.png" );

    $img->Set( magick => 'mono' );
    return( $img->ImageToBlob() );
}

sub createPlot {
    my $img = createImg();
    my $temps = getTempGraph();

    # draw 0°C line...
    $img->Draw(
        primitive => "line",
        'stroke-dasharray' => [2,6],
        fill => "none",
        stroke => "black",
        points => "0,".
            yscale( 0, $temps->{'low'}, $temps->{'high'} ).
            ",128,".
            yscale( 0, $temps->{'low'}, $temps->{'high'} )
    );

    for( my $i = 1; $i < @{$temps->{'temp'}}; $i++ ) {
        $img->Draw(
            primitive => "line",
            points => (128 - $i).",".
                yscale(
                    @{$temps->{'temp'}}[$i - 1],
                    $temps->{'low'},
                    $temps->{'high'} ).
                ",".(127 - $i).",".
                yscale(
                    @{$temps->{'temp'}}[$i],
                    $temps->{'low'},
                    $temps->{'high'})
        );
    }

    $img->Annotate(
        font => FONT,
        gravity => "north",
        geometry => "+0+0",
        size => "128x",
        fill => 'black',
        text => $temps->{'high'}." °C",
        pointsize => 16
    );

    $img->Annotate(
        font => FONT,
        gravity => "north",
        geometry => "+0+48",
        size => "128x",
        fill => 'black',
        text => $temps->{'low'}."°C",
        pointsize => 16
    );

    $img->Set( magick => 'mono' );

    return( $img->ImageToBlob() );
}
